import { RangeSetBuilder } from "@codemirror/state";
import { Decoration, EditorView, ViewPlugin } from "@codemirror/view";
import { getIndentUnit } from "@codemirror/language";
class IndentWrappedLinesPlugin {
    constructor(view, options) {
        this.initialPadding = null;
        this.view = view;
        this.indentUnit = getIndentUnit(view.state);
        this.options = options;
        this.isChrome = /Chrome/.test(navigator.userAgent);
        this.generate(view.state);
    }
    update(update) {
        const indentUnit = getIndentUnit(update.state);
        if (this.indentUnit !== indentUnit ||
            update.docChanged ||
            update.viewportChanged) {
            this.indentUnit = indentUnit;
            this.generate(update.state);
        }
    }
    destroy() {
    }
    generate(state) {
        const builder = new RangeSetBuilder();
        if (this.initialPadding) {
            this.appendStylesToBuilder(builder, state, this.initialPadding);
        }
        else {
            const read = (view) => {
                const element = view.contentDOM.querySelector(".cm-line");
                if (element) {
                    const initPadding = window.getComputedStyle(element).getPropertyValue("padding-left");
                    if (initPadding) {
                        this.initialPadding = initPadding;
                    }
                    else {
                        throw new Error("Could not get initial padding value");
                    }
                    this.appendStylesToBuilder(builder, state, this.initialPadding);
                }
            };
            this.view.requestMeasure({
                read,
            });
            this.decorations = builder.finish();
        }
        this.decorations = builder.finish();
    }
    appendStylesToBuilder(builder, state, initialPadding) {
        const lines = this.getVisibleLines(state);
        const initialIndentValue = this.getInitialIndentValue(state.tabSize);
        for (const line of lines) {
            const columnsDetails = this.getNumberOfColumns(line.text, state.tabSize);
            const paddingValue = `calc(${columnsDetails.amountOfSpacing + initialIndentValue}ch + ${initialPadding})`;
            let textIndentValue;
            if (this.isChrome) {
                textIndentValue = `calc(-${columnsDetails.amountOfSpacing + initialIndentValue}ch - ${Number(columnsDetails.containsTab)}px)`;
            }
            else {
                textIndentValue = `-${columnsDetails.amountOfSpacing + initialIndentValue}px`;
            }
            builder.add(line.from, line.from, Decoration.line({
                attributes: {
                    style: `padding-left: ${paddingValue}; text-indent: ${textIndentValue};`,
                },
            }));
        }
    }
    getVisibleLines(state) {
        const lines = new Set();
        let lastLine = undefined;
        for (const range of this.view.visibleRanges) {
            let position = range.from;
            while (position < range.to) {
                const line = state.doc.lineAt(position);
                if (line !== lastLine) {
                    lines.add(line);
                    lastLine = line;
                }
                position += line.length + 1;
            }
        }
        return lines;
    }
    getNumberOfColumns(value, tabSize) {
        let charSpacing = 0;
        let containsTab = false;
        for (let index = 0; index < value.length; index++) {
            if (value[index] === " ") {
                charSpacing++;
                continue;
            }
            else if (value[index] === "\t") {
                containsTab = true;
                charSpacing += tabSize - (charSpacing % tabSize);
                continue;
            }
            else if (value[index] === "\r") {
                continue;
            }
            else {
                break;
            }
        }
        return {
            amountOfSpacing: charSpacing,
            containsTab,
        };
    }
    getInitialIndentValue(tabSize) {
        switch (this.options.initialIndentType) {
            case "space": {
                return this.options.initialIndent;
            }
            case "tab": {
                return this.options.initialIndent * tabSize;
            }
            case "indentUnit": {
                return this.options.initialIndent * this.indentUnit;
            }
            default: {
                throw new Error(`Unknown initialIndentType: ${this.options.initialIndentType}`);
            }
        }
    }
}
/**
 * An extension for CodeMirror to indent wrapped lines
 * This extension also contains the extension to enable line wrapping
 *
 * @example
 * const extensions = [
 * 	indentWrappedLines(),
 * ];
 *
 * @example
 * const options = {
 * 	initialIndent: 4,
 * 	initialIndentType: "indentUnit",
 * };
 *
 * const extensions = [
 * 	indentWrappedLines(options),
 * ];
 *
 * @param options Options for the indent wrapped lines extension
 * @returns A CodeMirror extension
 */
export function indentWrappedLines(options) {
    const constructedOptions = Object.assign({ initialIndent: 0, initialIndentType: "space" }, options);
    class IndentWrappedLines extends IndentWrappedLinesPlugin {
        constructor(view) {
            super(view, constructedOptions);
        }
    }
    return [
        ViewPlugin.fromClass(IndentWrappedLines, {
            decorations: (view) => view.decorations,
        }),
        EditorView.lineWrapping,
    ];
}
